#ifndef _BTHANDLER_H_
#define _BTHANDLER_H_

#include "btutils.h"
#include "btsocket.h"

#include <bluetooth/bluetooth.h>
#include <bluetooth/sdp.h>
#include <bluetooth/sdp_lib.h>

struct search_context {
  char      *svc;    // Service
  uuid_t    group;   // Browse group
  int       tree;    // Display full attribute tree
  uint32_t  handle;  // Service record handle
};

/* Context information about the whole service */
struct service_context {
  struct uuid_def *     service;        /* Service UUID, if known */
};

struct sBTApp {
  int state;
  int pid;
  char param[BT_PPP_PARAM_LEN];
  struct sBTTimer removetr;
  struct sBTTimer ttytr;
  struct sBTTimer readytr;
  struct sBTTimer startretrytr;
  struct sBTTimer starttotaltr;
  int refcount;
};
typedef struct sBTApp BTApp;

struct sBTConnection {
  BTSocket sock;
  int profile;
  int flags;
  char pin[BT_PIN_LEN];
  int state;
  int error;
  //Buffer to keep data while connecting
  char* sendbuf;
  int sendbuflen;
  //Listen restart and total timeouts
  struct sBTTimer listenretrytr;
  struct sBTTimer listentotaltr;
  //Reconnect restart and total timeouts
  struct sBTTimer connectretrytr;
  struct sBTTimer connecttotaltr;
  //Bluetooth link removal timeout
  struct sBTTimer aclremovetr;
  //Pairing timers
  struct sBTTimer pairattempttr;
  struct sBTTimer pairtotaltr;
  //External app data
  BTApp ppp;
  //To speed up RIDER/GO connection times
  int reconnecting;
  //Dynamically retrieve channel
  int chan_retrive_pid;
  //Dynamically retrieve channel for dun
  int chan_retrive_pid_ppp;
  int sdpnrretries;
  //To manage connection requests fake connection record is used
  int fake;
  // hcitool: bremen hack
  int hcitool_pid;
  int hcitoolnrretries;

};
typedef struct sBTConnection *BTConnection;
typedef const struct sBTConnection *cBTConnection;

int BTHandlerCompare(cBTConnection con, const char *bdaddr, int chan, int profile, int *res);

int BTHandlerCreate(BTConnection *con, const char *bdaddr, int chan, int profile, int flags, const char *pin);
int BTHandlerDestroy(BTConnection con);
int BTHandlerCreateRequest(BTConnection *con);

int BTHandlerStartApp(BTConnection con, const char *param);
int BTHandlerStopApp(BTConnection con);

int BTHandlerSend(BTConnection con, const char *buf, int len, int *sent);
int BTHandlerReceive(BTConnection con, char *buf, int len, int *received);

int BTHandlerRelease(BTConnection con);
int BTHandlerClose(BTConnection con);

int BTHandlerGetState(cBTConnection con, int *state);
int BTHandlerGetError(cBTConnection con, int *error);

int BTHandlerSetFlags(BTConnection con, int flags);
int BTHandlerGetFlags(cBTConnection con, int *flags);

int BTHandlerIsProfileActive(cBTConnection con, int profile, int *res);
int BTHandlerNeedPair(cBTConnection con, int *res);

int BTHandlerRun(BTConnection con, int disallow_connect);

int BTHandlerCreateConnectionRequest(BTConnection *con);
int BTHandlerRequestConnection(BTConnection con);
int BTHandlerReleaseConnectionRequest(BTConnection con);
int BTHandlerReconnect(BTConnection con);

int BTHandlerGetBTName(char* name, int size);
int BTHandlerSetBTName(char* name);
int BTHandlerExitSniffMode(char* bdaddr);

int BTHandlerScanProfileCH(const char *bAdress, int SearchProfile);

int BTHandlerGetMtuSize(cBTConnection con, int * imtu, int * omtu);
int BTHandlerCreateTTY(cBTConnection con, int tty_id);
int BTHandlerReleaseTTY(cBTConnection con);


#endif // _BTHANDLER_H_
