#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include "btserverdefines.h"
#include "btutils.h"
#include "btttmp.h"
#include "bthandler.h"
#include "bthidremote.h"
#include "btserver.h"
//PLEASECHECK: following include is needed for the hack down there, remove when fixed
#include "TTMPConn.h"
#include "bthidremote.h"
#include "hidd.h"

//this is returned by a TTMP macro if communication error occurs (undefine == return;)
#undef TTMP_ERROR
#define TTMP_ERROR

static int serverbusy = 1;
static char serverbuf[BTSR_BUFFER_SIZE];
static TTMPConnection ttmpcon[BT_MAX_CLIENTS];
static int ttmppend[BT_MAX_CLIENTS];
static BTConnection btcon[BT_MAX_HANDLES];
static int refcon[BT_MAX_HANDLES][BT_MAX_CLIENTS];
/* Assume pairing is disabled by default */
static int pairingenabled = 0;
/* PLEASECHECK: according to Johan, there is no need to disable/enable pagescan anymore */
#undef HANDLE_PAGESCAN
#ifdef HANDLE_PAGESCAN
static int pagescanenabled = 0;
/* PLEASECHECK:  This should be removed when the kernel part is fixed */
static int pagescanenablepid = -1;
static int pagescandisablepid = -1;
#endif /* HANDLE_PAGESCAN */

#ifdef ENABLE_DEBUG_LOG

#define STATE2STR(state) state2str(state)
#define ERROR2STR(error) error2str(error)

#include <stdio.h>
#include <signal.h>
#include <errno.h>
#include <ucontext.h>


static void sighandler_usr2()
{
  int i, j, res;
  printf("****************** btserver state begin ************************\n");
  res = 0;
  for (i = 0; i < BT_MAX_CLIENTS; i++)
    if (ttmpcon[i])
      res++;
  printf("The number of active clients: %d\n", res);
  res = 0;
  for (i = 0; i < BT_MAX_HANDLES; i++)
    if (btcon[i])
      res++;
  printf("The number of active handles: %d\n", res);
  printf("Reference map:\n");
  printf("Handles/Clients ");
  for (i = 0; i < BT_MAX_CLIENTS; i++)
    printf("%2d ", i);
  printf("State");
  printf("\n");
  for (i = 0; i < BT_MAX_HANDLES; i++) {
    printf("    %2d          ", i);
    for (j = 0; j < BT_MAX_CLIENTS; j++)
      printf("%2d ", refcon[i][j]);
    if (btcon[i])
      printf("%s\n", STATE2STR(btcon[i]->state));
    else
      printf("-\n");
  }
  printf("\n");
  printf("Pairing is %s.\n", pairingenabled?"enabled":"disabled");
#ifdef HANDLE_PAGESCAN
  printf("Page scan is %s.\n", pagescanenabled?"enabled":"disabled");
#endif /* HANDLE_PAGESCAN */
  printf("******************* btserver state end *************************\n");
}

#else //ENABLE_DEBUG_LOG

#define STATE2STR(state) "DEBUG DISABLED"
#define ERROR2STR(error) "DEBUG DISABLED"

#endif //ENABLE_DEBUG_LOG

inline int refhandle(int handle)
{
  int i, res = 0;
  for (i = 0; i < BT_MAX_CLIENTS; i++)
    res += refcon[handle][i];
  return res;
}

void requesthandler(int ittmpcon, TTMPMessage msg)
{
  int i, cmd, handle, res, cmdid;

  TTMP_GETFIELDINT(msg, CMD_ID, &cmdid);
  TTMP_GETFIELDINT(msg, CMD_PAR_COMMAND, &cmd);

  switch(cmd)
  {
  case CMD_INIT:
  {
    char *bdaddr = 0, *pin = 0;
    int chan, prof, opt;

    TTMP_DETACHFIELDSTR(msg, CMD_PAR_ADDRESS, &bdaddr);
    TTMP_GETFIELDINT(msg, CMD_PAR_CHANNEL, &chan);
    TTMP_GETFIELDINT(msg, CMD_PAR_PROFILE, &prof);
    TTMP_GETFIELDINT(msg, CMD_PAR_OPTION, &opt);
    TTMP_DETACHFIELDSTR(msg, CMD_PAR_PIN, &pin);
    if (!bdaddr || !pin) {
      DEBUG_LOG("Failed to retrive bdaddr or pin from TTMP");
      handle = -1;
      goto init_failed;
    }
    for (i = 0; i < BT_MAX_HANDLES; i++) {
      if (!btcon[i] || btcon[i]->fake)
        continue;
      if (BTHandlerCompare(btcon[i], bdaddr, chan, prof, &res) != BT_ERR_NONE)
        continue;
      if (res == BT_TRUE)
        break;
    }
    if (i < BT_MAX_HANDLES)
      handle = i;
    else {
      for (i = 0; i < BT_MAX_HANDLES; i++)
        if (!btcon[i])
          break;
      if (i == BT_MAX_HANDLES)
        handle = -1;
      else if (BTHandlerCreate(&btcon[i], bdaddr, chan, prof, opt, pin) != BT_ERR_NONE)
        handle = -1;
      else
        handle = i;
    }
    if (handle != -1)
      refcon[handle][ittmpcon]++;
init_failed:
    ttmppend[ittmpcon]++;
    if (bdaddr)
      free(bdaddr);
    if (pin)
      free(pin);
    TTMP_CLEAR(msg);
    TTMP_SETFIELDINT(msg, CMD_ID, cmdid);
    TTMP_SETFIELDINT(msg, CMD_PAR_HANDLE, handle);
    TTMP_SEND(ttmpcon[ittmpcon], msg);
    ttmppend[ittmpcon]--;
    if (handle != -1) {
      DEBUG_LOG3("CMD_INIT: returned client=%d, handle=%d, refcount=%d", ittmpcon, handle, refhandle(handle));
    } else {
      DEBUG_LOG2("CMD_INIT: returned client=%d, handle=%d", ittmpcon, handle);
    }
    break;
  }
  case CMD_CLOSE:
  {
    TTMP_GETFIELDINT(msg, CMD_PAR_HANDLE, &handle);
    if (handle < BT_MAX_HANDLES && handle > -1) {
      if (btcon[handle]) {
        DEBUG_LOG3("CMD_CLOSE: client=%d, handle=%d, refcount=%d", ittmpcon, handle, refhandle(handle));
        refcon[handle][ittmpcon]--;
        if (!refhandle(handle))
          res = BTHandlerClose(btcon[handle]);
        else
          res = BT_ERR_NONE;
      }
      else
        res = BT_ERR_ARGUMENT;
    }
    else
      res = BT_ERR_ARGUMENT;
    ttmppend[ittmpcon]++;
    TTMP_CLEAR(msg);
    TTMP_SETFIELDINT(msg, CMD_ID, cmdid);
    TTMP_SETFIELDINT(msg, CMD_PAR_RESULT, res);
    TTMP_SEND(ttmpcon[ittmpcon], msg);
    ttmppend[ittmpcon]--;

    break;
  }
  case CMD_RELEASE:
  {
    TTMP_GETFIELDINT(msg, CMD_PAR_HANDLE, &handle);
    if (handle < BT_MAX_HANDLES && handle > -1) {
      if (btcon[handle]) {
        DEBUG_LOG2("CMD_RELEASE: client=%d, handle=%d", ittmpcon, handle);
        res = BTHandlerRelease(btcon[handle]);
      }
      else
        res = BT_ERR_ARGUMENT;
    }
    else
      res = BT_ERR_ARGUMENT;
    ttmppend[ittmpcon]++;
    TTMP_CLEAR(msg);
    TTMP_SETFIELDINT(msg, CMD_ID, cmdid);
    TTMP_SETFIELDINT(msg, CMD_PAR_RESULT, res);
    TTMP_SEND(ttmpcon[ittmpcon], msg);
    ttmppend[ittmpcon]--;

    break;
  }
  case CMD_SEND:
  {
    int len, sent;
    char* buf = 0;

    TTMP_GETFIELDINT(msg, CMD_PAR_HANDLE, &handle);
    TTMP_GETFIELDSIZE(msg, CMD_PAR_DATA, &len);
    TTMP_DETACHFIELDBIN(msg, CMD_PAR_DATA, &buf);
    if (!buf) {
      DEBUG_LOG("Failed to retrive buf from TTMP");
      res = BT_ERR_FAILED;
      goto send_failed;
    }
    if (handle < BT_MAX_HANDLES && handle > -1) {
      if (btcon[handle]) {
        //				DEBUG_LOG3("CMD_SEND: client=%d, handle=%d, len=%d", ittmpcon, handle, len);
        res = BTHandlerSend(btcon[handle], buf, len, &sent);
      }
      else
        res = BT_ERR_ARGUMENT;
    }
    else
      res = BT_ERR_ARGUMENT;
    if (buf)
      free(buf);
send_failed:
    ttmppend[ittmpcon]++;
    TTMP_CLEAR(msg);
    TTMP_SETFIELDINT(msg, CMD_ID, cmdid);
    TTMP_SETFIELDINT(msg, CMD_PAR_RESULT, res);
    TTMP_SEND(ttmpcon[ittmpcon], msg);
    ttmppend[ittmpcon]--;

    break;
  }
  case CMD_RECEIVE:
  {
    int len, received;

    TTMP_GETFIELDINT(msg, CMD_PAR_HANDLE, &handle);
    TTMP_GETFIELDINT(msg, CMD_PAR_SIZE, &len);
    if (len > BTSR_BUFFER_SIZE)
      len = BTSR_BUFFER_SIZE;
    if (handle < BT_MAX_HANDLES && handle > -1) {
      if (btcon[handle]) {
        //				DEBUG_LOG3("CMD_RECEIVE: client=%d, handle=%d, len=%d", ittmpcon, handle, len);
        res = BTHandlerReceive(btcon[handle], serverbuf, len, &received);
      }
      else
        res = BT_ERR_ARGUMENT;
    }
    else
      res = BT_ERR_ARGUMENT;
    ttmppend[ittmpcon]++;
    TTMP_CLEAR(msg);
    TTMP_SETFIELDINT(msg, CMD_ID, cmdid);
    TTMP_SETFIELDINT(msg, CMD_PAR_RESULT, res);
    if (res != BT_ERR_NONE)
      received = 0;
    TTMP_SETFIELDINT(msg, CMD_PAR_SIZE, received);
    if (received)
      TTMP_SETFIELDBIN(msg, CMD_PAR_DATA, serverbuf, received);
    TTMP_SEND(ttmpcon[ittmpcon], msg);
    ttmppend[ittmpcon]--;

    break;
  }
  case CMD_GETSTATE:
  {
    int state;
    TTMP_GETFIELDINT(msg, CMD_PAR_HANDLE, &handle);
    if (handle < BT_MAX_HANDLES && handle > -1) {
      if (btcon[handle]) {
        //				DEBUG_LOG2("CMD_GETSTATE: client=%d, handle=%d", ittmpcon, handle);
        if (BTHandlerGetState(btcon[handle], &state) != BT_ERR_NONE)
          state = BT_STATE_UNINITIALIZED;
      }
      else
        state = BT_STATE_UNINITIALIZED;
    }
    else
      state = BT_STATE_UNINITIALIZED;
    ttmppend[ittmpcon]++;
    TTMP_CLEAR(msg);
    TTMP_SETFIELDINT(msg, CMD_ID, cmdid);
    TTMP_SETFIELDINT(msg, CMD_PAR_RESULT, state);
    TTMP_SEND(ttmpcon[ittmpcon], msg);
    ttmppend[ittmpcon]--;

    break;
  }
  case CMD_GETERROR:
  {
    TTMP_GETFIELDINT(msg, CMD_PAR_HANDLE, &handle);
    if (handle < BT_MAX_HANDLES && handle > -1) {
      if (btcon[handle]) {
        DEBUG_LOG2("CMD_ERROR: client=%d, handle=%d", ittmpcon, handle);
        if (BTHandlerGetError(btcon[handle], &res) != BT_ERR_NONE)
          res = BT_ERR_ARGUMENT;
      }
      else
        res = BT_ERR_ARGUMENT;
    }
    else
      res = BT_ERR_ARGUMENT;
    ttmppend[ittmpcon]++;
    TTMP_CLEAR(msg);
    TTMP_SETFIELDINT(msg, CMD_ID, cmdid);
    TTMP_SETFIELDINT(msg, CMD_PAR_RESULT, res);
    TTMP_SEND(ttmpcon[ittmpcon], msg);
    ttmppend[ittmpcon]--;

    break;
  }
  case CMD_STARTUSINGCHANNEL:
  {
    char* buf = 0;

    TTMP_GETFIELDINT(msg, CMD_PAR_HANDLE, &handle);
    TTMP_DETACHFIELDSTR(msg, CMD_PAR_DATA, &buf);
    if (!buf) {
      DEBUG_LOG("Failed to retrive buf from TTMP");
      res = BT_ERR_FAILED;
      goto startusing_failed;
    }
    if (handle < BT_MAX_HANDLES && handle > -1) {
      if (btcon[handle]) {
        DEBUG_LOG2("CMD_STARTUSINGCHANNEL: client=%d, handle=%d", ittmpcon, handle);
        res = BTHandlerStartApp(btcon[handle], buf);
      }
      else
        res = BT_ERR_ARGUMENT;
    }
    else
      res = BT_ERR_ARGUMENT;
    if (buf)
      free(buf);
startusing_failed:
    ttmppend[ittmpcon]++;
    TTMP_CLEAR(msg);
    TTMP_SETFIELDINT(msg, CMD_ID, cmdid);
    TTMP_SETFIELDINT(msg, CMD_PAR_RESULT, res);
    TTMP_SEND(ttmpcon[ittmpcon], msg);
    ttmppend[ittmpcon]--;

    break;
  }
  case CMD_STOPUSINGCHANNEL:
  {
    TTMP_GETFIELDINT(msg, CMD_PAR_HANDLE, &handle);
    if (handle < BT_MAX_HANDLES && handle > -1) {
      if (btcon[handle]) {
        DEBUG_LOG2("CMD_STOPUSINGCHANNEL: client=%d, handle=%d", ittmpcon, handle);
        res = BTHandlerStopApp(btcon[handle]);
      }
      else
        res = BT_ERR_ARGUMENT;
    }
    else
      res = BT_ERR_ARGUMENT;
    ttmppend[ittmpcon]++;
    TTMP_CLEAR(msg);
    TTMP_SETFIELDINT(msg, CMD_ID, cmdid);
    TTMP_SETFIELDINT(msg, CMD_PAR_RESULT, res);
    TTMP_SEND(ttmpcon[ittmpcon], msg);
    ttmppend[ittmpcon]--;

    break;
  }
  case CMD_GETFLAGS:
  {
    int flags;

    TTMP_GETFIELDINT(msg, CMD_PAR_HANDLE, &handle);
    if (handle < BT_MAX_HANDLES && handle > -1) {
      if (btcon[handle]) {
        DEBUG_LOG2("CMD_GETFLAGS: client=%d, handle=%d", ittmpcon, handle);
        if (BTHandlerGetFlags(btcon[handle], &flags) != BT_ERR_NONE)
          flags = BT_FLAG_UNDEFINED;
      }
      else
        flags = BT_FLAG_UNDEFINED;
    }
    else
      flags = BT_FLAG_UNDEFINED;
    ttmppend[ittmpcon]++;
    TTMP_CLEAR(msg);
    TTMP_SETFIELDINT(msg, CMD_ID, cmdid);
    TTMP_SETFIELDINT(msg, CMD_PAR_RESULT, flags);
    TTMP_SEND(ttmpcon[ittmpcon], msg);
    ttmppend[ittmpcon]--;

    break;
  }
  case CMD_SETFLAGS:
  {
    int flags;

    TTMP_GETFIELDINT(msg, CMD_PAR_HANDLE, &handle);
    TTMP_GETFIELDINT(msg, CMD_PAR_OPTION, &flags);
    if (handle < BT_MAX_HANDLES && handle > -1) {
      if (btcon[handle]) {
        DEBUG_LOG2("CMD_SETFLAGS: client=%d, handle=%d", ittmpcon, handle);
        res = BTHandlerSetFlags(btcon[handle], flags);
      }
      else
        res = BT_ERR_ARGUMENT;
    }
    else
      res = BT_ERR_ARGUMENT;
    ttmppend[ittmpcon]++;
    TTMP_CLEAR(msg);
    TTMP_SETFIELDINT(msg, CMD_ID, cmdid);
    TTMP_SETFIELDINT(msg, CMD_PAR_RESULT, res);
    TTMP_SEND(ttmpcon[ittmpcon], msg);
    ttmppend[ittmpcon]--;

    break;
  }
  case CMD_ISPROFILEACTIVE:
  {
    int profile, active;

    TTMP_GETFIELDINT(msg, CMD_PAR_PROFILE, &profile);
    DEBUG_LOG2("CMD_ISPROFILEACTIVE: client=%d, handle=%d", ittmpcon, handle);
    active = BT_FALSE;
    for (i = 0; i < BT_MAX_HANDLES; i++)
      if (btcon[i]) {
        if (BTHandlerIsProfileActive(btcon[i], profile, &active) == BT_ERR_NONE) {
          if (active == BT_TRUE)
            break;
        } else {
          active = BT_FALSE;
          break;
        }
      }
    ttmppend[ittmpcon]++;
    TTMP_CLEAR(msg);
    TTMP_SETFIELDINT(msg, CMD_ID, cmdid);
    TTMP_SETFIELDINT(msg, CMD_PAR_RESULT, active);
    TTMP_SEND(ttmpcon[ittmpcon], msg);
    ttmppend[ittmpcon]--;

    break;
  }
  case CMD_REQUESTCONNECTION:
  {
    for (i = 0; i < BT_MAX_HANDLES; i++)
      if (!btcon[i])
        break;
    if (i == BT_MAX_HANDLES)
      handle = -1;
    else if (BTHandlerCreateRequest(&btcon[i]) != BT_ERR_NONE)
      handle = -1;
    else
      handle = i;
    if (handle != -1)
      refcon[handle][ittmpcon]++;

    ttmppend[ittmpcon]++;
    TTMP_CLEAR(msg);
    TTMP_SETFIELDINT(msg, CMD_ID, cmdid);
    TTMP_SETFIELDINT(msg, CMD_PAR_HANDLE, handle);
    TTMP_SEND(ttmpcon[ittmpcon], msg);
    ttmppend[ittmpcon]--;
    if (handle != -1) {
      DEBUG_LOG2("CMD_REQUESTCONNECTION: returned client=%d, handle=%d", ittmpcon, handle);
    } else {
      DEBUG_LOG1("CMD_REQUESTCONNECTION: FAILED client=%d", ittmpcon);
    }
    break;
  }
  case CMD_RELEASECONNECTIONREQUEST:
  {
    TTMP_GETFIELDINT(msg, CMD_PAR_HANDLE, &handle);

    if (handle < BT_MAX_HANDLES && handle > -1) {
      if (btcon[handle]) {
        DEBUG_LOG2("CMD_RELEASECONNECTIONREQUEST: client=%d, handle=%d", ittmpcon, handle);
        BTHandlerDestroy(btcon[handle]);
        btcon[handle] = NULL;
        refcon[handle][ittmpcon]--;
      }
      else
        res = BT_ERR_ARGUMENT;
    }
    else
      res = BT_ERR_ARGUMENT;

    ttmppend[ittmpcon]++;
    TTMP_CLEAR(msg);
    TTMP_SETFIELDINT(msg, CMD_ID, cmdid);
    TTMP_SEND(ttmpcon[ittmpcon], msg);
    ttmppend[ittmpcon]--;

    break;
  }
  case CMD_GETBTNAME:
  {
    res = BTHandlerGetBTName(serverbuf, BTSR_BUFFER_SIZE);
    ttmppend[ittmpcon]++;
    TTMP_CLEAR(msg);
    TTMP_SETFIELDINT(msg, CMD_ID, cmdid);
    if (res == BT_ERR_NONE) {
      TTMP_SETFIELDSTR(msg, CMD_PAR_DATA, serverbuf);
      TTMP_SETFIELDINT(msg, CMD_PAR_SIZE, strlen(serverbuf));
    }
    TTMP_SETFIELDINT(msg, CMD_PAR_RESULT, res);
    TTMP_SEND(ttmpcon[ittmpcon], msg);
    ttmppend[ittmpcon]--;

    break;
  }
  case CMD_SETBTNAME:
  {
    char* buf = 0;

    TTMP_DETACHFIELDSTR(msg, CMD_PAR_DATA, &buf);
    if (!buf) {
      DEBUG_LOG("Failed to retrive buf from TTMP");
      res = BT_ERR_FAILED;
      goto setbtname_failed;
    }
    res = BTHandlerSetBTName(buf);
    if (buf)
      free(buf);
setbtname_failed:
    ttmppend[ittmpcon]++;
    TTMP_CLEAR(msg);
    TTMP_SETFIELDINT(msg, CMD_ID, cmdid);
    TTMP_SETFIELDINT(msg, CMD_PAR_RESULT, res);
    TTMP_SEND(ttmpcon[ittmpcon], msg);
    ttmppend[ittmpcon]--;

    break;
  }
#ifdef BT_HID_REMOTE
  case CMD_REQUEST_HID_CONN:
  {
    // DEBUG_LOG("btserver: Received CMD_REQUEST_HID_CONN\n");
    char *name = NULL;
    char *address = NULL;

    ttmppend[ittmpcon]++;
    TTMP_CLEAR(msg);
    TTMP_SETFIELDINT(msg, CMD_ID, cmdid);
    TTMP_SETFIELDINT(msg, CMD_PAR_COMMAND, CMD_REPLY_HID_CONN);

    name = getHidRemoteName();
    address = getHidRemoteAddress();

    if( name != NULL )
    {
      TTMP_SETFIELDSTR(msg, CMD_PAR_DATA, address);
      TTMP_SETFIELDINT(msg, CMD_PAR_SIZE, strlen(address));
      DEBUG_LOG2("btserver: name:%s address:%s\n", name, address);
      res = BT_ERR_NONE;
    }
    else
    {
      TTMP_SETFIELDSTR(msg, CMD_PAR_DATA, "nothing");
      TTMP_SETFIELDINT(msg, CMD_PAR_SIZE, strlen("nothing"));
      res = BT_ERR_FAILED;
    }
    TTMP_SETFIELDINT(msg, CMD_PAR_RESULT, res);

    TTMP_SEND(ttmpcon[ittmpcon], msg);
    ttmppend[ittmpcon]--;

    break;
  }
  case CMD_CONNECT_HID:
  {
    char address[248];
    int result = BT_ERR_NONE;

    TTMP_GETFIELDBIN(msg, CMD_PAR_DATA, address, 248);
    DEBUG_LOG2("btserver: CMD_CONNECT_HID: client=%d, address=%s", ittmpcon, address);

    ttmppend[ittmpcon]++;
    TTMP_CLEAR(msg);
    TTMP_SETFIELDINT(msg, CMD_ID, cmdid);
    TTMP_SETFIELDINT(msg, CMD_PAR_COMMAND, CMD_REPLY_CONNECT_HID);

    // Do not check input address to current address
    result = connect_device();

    TTMP_SETFIELDINT(msg, CMD_PAR_RESULT, result);

    TTMP_SEND(ttmpcon[ittmpcon], msg);
    ttmppend[ittmpcon]--;

    break;
  }
  case CMD_START_HID:
  {
    int result = BT_ERR_NONE;

    DEBUG_LOG1("btserver: CMD_START_HID: client=%d", ittmpcon);

    ttmppend[ittmpcon]++;
    TTMP_CLEAR(msg);
    TTMP_SETFIELDINT(msg, CMD_ID, cmdid);
    TTMP_SETFIELDINT(msg, CMD_PAR_COMMAND, CMD_REPLY_START_HID);

    result = start_hid();

    TTMP_SETFIELDINT(msg, CMD_PAR_RESULT, result);

    TTMP_SEND(ttmpcon[ittmpcon], msg);
    ttmppend[ittmpcon]--;

    break;
  }

  case CMD_DISCONNECT_HID:
  {
    int result = BT_ERR_NONE;

    DEBUG_LOG1("btserver: CMD_DISCONNECT_HID: client=%d", ittmpcon);

    ttmppend[ittmpcon]++;
    TTMP_CLEAR(msg);
    TTMP_SETFIELDINT(msg, CMD_ID, cmdid);
    TTMP_SETFIELDINT(msg, CMD_PAR_COMMAND, CMD_REPLY_DISCONNECT_HID);

    result = disconnect_device();

    TTMP_SETFIELDINT(msg, CMD_PAR_RESULT, result);

    TTMP_SEND(ttmpcon[ittmpcon], msg);
    ttmppend[ittmpcon]--;

    break;
  }
  
#endif // BT_HID_REMOTE

  case CMD_GET_MTU_SIZE:
  {
    int imtu=-1, omtu=-1;
    int res = 0;

    TTMP_GETFIELDINT(msg, CMD_PAR_HANDLE, &handle);
    if (handle < BT_MAX_HANDLES && handle > -1) {
      if (btcon[handle]) {
        DEBUG_LOG2("btserver: CMD_GET_MTU_SIZE: client=%d, handle=%d", ittmpcon, handle);
        if (BTHandlerGetMtuSize(btcon[handle], &imtu, &omtu) == BT_ERR_NONE)
        {
          res = 1;	// success
        }
        else
        {
          // failed to retreive mtu's
        }
      }
      else
      {
         // not a valid connection
      }
    }
    else
    {
      // not a valid handle
    }
    ttmppend[ittmpcon]++;
    TTMP_CLEAR(msg);
    TTMP_SETFIELDINT(msg, CMD_ID, cmdid);
    TTMP_SETFIELDINT(msg, CMD_PAR_RESULT, res);
    TTMP_SETFIELDINT(msg, CMD_PAR_DATA, imtu);  // first mtu-item is named as 'data'
    TTMP_SETFIELDINT(msg, CMD_PAR_SIZE, omtu);  // second mtu-item is names as 'size'
    TTMP_SEND(ttmpcon[ittmpcon], msg);
    ttmppend[ittmpcon]--;

    break;
  }
  
  case CMD_CREATE_TTY:
  {
    int tty_id=-1;
    int res = 0;

    TTMP_GETFIELDINT(msg, CMD_PAR_HANDLE, &handle);
    TTMP_GETFIELDINT(msg, CMD_PAR_DATA, &tty_id);

    DEBUG_LOG3("CMD_CREATE_TTY:handle=%d, clientid=%d, port=%d, \n", handle, ittmpcon, tty_id);

    if (handle < BT_MAX_HANDLES && handle > -1) {
      if (btcon[handle]) {
        if ((res = BTHandlerCreateTTY(btcon[handle], tty_id)) == BT_ERR_NONE)
        {
          res = 1;	// success
        }
      }
     }

    ttmppend[ittmpcon]++;
    TTMP_CLEAR(msg);
    TTMP_SETFIELDINT(msg, CMD_ID, cmdid);
    TTMP_SETFIELDINT(msg, CMD_PAR_RESULT, res);
    TTMP_SEND(ttmpcon[ittmpcon], msg);
    ttmppend[ittmpcon]--;
    break;
  }
  
  case CMD_RELEASE_TTY:
  {
    int res = 0;

    TTMP_GETFIELDINT(msg, CMD_PAR_HANDLE, &handle);

    DEBUG_LOG2("CMD_RELEASE_TTY:handle=%d, clientid=%d\n", handle, ittmpcon);

    if (handle < BT_MAX_HANDLES && handle > -1) {
      if (btcon[handle]) {
        if ((res = BTHandlerReleaseTTY(btcon[handle])) == BT_ERR_NONE)
        {
          res = 1;	// success
        }
      }
     }

    ttmppend[ittmpcon]++;
    TTMP_CLEAR(msg);
    TTMP_SETFIELDINT(msg, CMD_ID, cmdid);
    TTMP_SETFIELDINT(msg, CMD_PAR_RESULT, res);
    TTMP_SEND(ttmpcon[ittmpcon], msg);
    ttmppend[ittmpcon]--;
    break;
  }
  
  case CMD_RECONNECT:
  {
    TTMP_GETFIELDINT(msg, CMD_PAR_HANDLE, &handle);
    if (handle < BT_MAX_HANDLES && handle > -1) {
      if (btcon[handle]) {
        DEBUG_LOG2("CMD_RECONNECT: client=%d, handle=%d", ittmpcon, handle);
        res = BTHandlerReconnect(btcon[handle]);
      }
      else
        res = BT_ERR_ARGUMENT;
    }
    else
      res = BT_ERR_ARGUMENT;
    ttmppend[ittmpcon]++;
    TTMP_CLEAR(msg);
    TTMP_SETFIELDINT(msg, CMD_ID, cmdid);
    TTMP_SETFIELDINT(msg, CMD_PAR_RESULT, res);
    TTMP_SEND(ttmpcon[ittmpcon], msg);
    ttmppend[ittmpcon]--;

    break;
  }

  case CMD_EXIT_SNIFF_MODE:
  { 
    char *bdaddr = 0;
    res = BT_ERR_ARGUMENT;
    TTMP_DETACHFIELDSTR(msg, CMD_PAR_ADDRESS, &bdaddr);
    DEBUG_LOG1("btserver::CMD_EXIT_SNIFF_MODE. bdaddr= %s", bdaddr);
    
    if (!bdaddr) {
      DEBUG_LOG("Failed to retrive bdaddr from TTMP");
      res = BT_ERR_ARGUMENT;
      goto bdaddr_failed;
    }    
    if (BTHandlerExitSniffMode(bdaddr)==0){  
       DEBUG_LOG("btserver:BTHandlerExitSniffMode(bdaddr)==0 | succeeded");
       res = BT_ERR_NONE;
    } 
    else{  
       DEBUG_LOG("btserver:BTHandlerExitSniffMode(bdaddr)==0 | failed");
       res = BT_ERR_FAILED;   
    } 
              
    if (bdaddr)
      free(bdaddr); 
      
    bdaddr_failed:     
      ttmppend[ittmpcon]++;  
      TTMP_CLEAR(msg);
      TTMP_SETFIELDINT(msg, CMD_ID, cmdid);
      TTMP_SETFIELDINT(msg, CMD_PAR_RESULT, res);
      TTMP_SEND(ttmpcon[ittmpcon], msg);
      ttmppend[ittmpcon]--;
      DEBUG_LOG1("btserver::CMD_EXIT_SNIFF_MODE. Returned %d.", res);
      break;
  }
  case CMD_GETBTCONNECTIONINFO:
  {
    char bdaddrstr[BT_BDADDR_LEN];
    
    TTMP_GETFIELDINT(msg, CMD_PAR_HANDLE, &handle);
    if (handle < BT_MAX_HANDLES && handle > -1) 
    {
      if (btcon[handle]) 
      {
        DEBUG_LOG3("CMD_GETBTCONNECTIONINFO: client=%d, handle=%d bdaddr lenght=%d", ittmpcon, handle, BT_BDADDR_LEN);
        res = BTSocketGetAddress(btcon[handle]->sock, bdaddrstr);
        DEBUG_LOG1("CMD_GETBTCONNECTIONINFO: BTSocketGetAddress returns addr=%s", bdaddrstr);
        DEBUG_LOG1("CMD_GETBTCONNECTIONINFO: addr strlen=%d", strlen(bdaddrstr));
      }
      else
      {
        DEBUG_LOG("CMD_GETBTCONNECTIONINFO: no btcon[handle]");
        res = BT_ERR_ARGUMENT;
      }
    }
    else
    {
      DEBUG_LOG("CMD_GETBTCONNECTIONINFO: no handle");
      res = BT_ERR_ARGUMENT;
    }  
    ttmppend[ittmpcon]++;
    TTMP_CLEAR(msg);
    TTMP_SETFIELDINT(msg, CMD_ID, cmdid);
    TTMP_SETFIELDINT(msg, CMD_PAR_RESULT, res);
    if (res == BT_ERR_NONE) 
    {
      TTMP_SETFIELDSTR(msg, CMD_PAR_ADDRESS, bdaddrstr);
      DEBUG_LOG("CMD_GETBTCONNECTIONINFO: sending the address");
    }
    TTMP_SEND(ttmpcon[ittmpcon], msg);
    ttmppend[ittmpcon]--;
    break;
  }
  default:
    break;
  }
}

int main(int argc, char** argv)
{
  TTMPServer svr;
  TTMPConnection newttmpcon;
  TTMPMessage msg;
  int i, j, res, state;
  int disallow_connect;
#ifdef ENABLE_DEBUG_LOG
  struct sigaction sigact;
#endif //ENABLE_DEBUG_LOG
  /*int returnValue = 0;*/

  if (argc > 1) {
    printf("Use: %s  (creates socket at [%s])\n", argv[0], SERVER_SOCKET_ADDR);
    exit(1);
  }
  signal(SIGPIPE, SIG_IGN);
  unlink(SERVER_SOCKET_ADDR);
  res = TTMPServerCreate(&svr, 0, BTSR_VER_MAJOR, BTSR_VER_MINOR);
  if (res < TTMP_ERR_NONE) {
    DEBUG_LOG1("Unable to create server, error %d", res);
    exit(1);
  }
  res = TTMPServerListen(svr, SERVER_SOCKET_ADDR);
  if (res < TTMP_ERR_NONE) {
    DEBUG_LOG1("Unable to listen, error %d", res);
    TTMPServerDestroy(svr);
    exit(1);
  }
  for (i = 0; i < BT_MAX_CLIENTS; i++) {
    ttmpcon[i] = NULL;
    ttmppend[i] = 0;
  }
  for (i = 0; i < BT_MAX_HANDLES; i++) {
    btcon[i] = NULL;
    for (j = 0; j < BT_MAX_CLIENTS; j++)
      refcon[i][j] = 0;
  }

#ifdef BT_HID_REMOTE
  BTHidRemoteHandleInit();
  int hidargc = 0;
  char *hidargv[3];
  char name[12] = "./hidd";
  char server[12] = "--server";
  char daemon[12] = "-n";
  hidargv[hidargc] = name; hidargc++;
  hidargv[hidargc] = daemon; hidargc++;
  hidargv[hidargc] = server; hidargc++;

#ifdef ENABLE_DEBUG_LOG
  for( i = 0; i < hidargc; i++)
  {
    DEBUG_LOG2("hidd argc:%d:%s\n", i, hidargv[i]);
  }
#endif // ENABLE_DEBUG_LOG
  /* returnValue = */ hidd(hidargc, hidargv);
#endif // BT_HID_REMOTE

#ifdef ENABLE_DEBUG_LOG
  /* Setup signal handler, used to dump the state on USR1 */
  sigact.sa_flags = SA_SIGINFO;
  sigact.sa_sigaction = (void *)sighandler_usr2;
  if(sigaction (SIGUSR2, &sigact, (struct sigaction *)NULL))
    DEBUG_LOG("could not register a signal handler");
#endif //ENABLE_DEBUG_LOG
  while(serverbusy) {

#ifdef BT_HID_REMOTE
    BTHidRemoteHandle();
#endif // BT_HID_REMOTE

    res = TTMPServerAccept(svr, &newttmpcon);
    if (res < TTMP_ERR_NONE) {
      if (res != TTMP_ERR_TIMEOUT)
        DEBUG_LOG("Incompatible client.");
    } else {
      //allocate new connection
      for( i = 0; i < BT_MAX_CLIENTS; i++)
        if (!ttmpcon[i])
          break;
      if (i == BT_MAX_CLIENTS) {
        DEBUG_LOG("Max connections reached");
        TTMPConnectionDestroy(newttmpcon);
      } else {
        DEBUG_LOG1("New client has been added with id %d", i);
        ttmpcon[i] = newttmpcon;
      }
    }
    //PLEASECHECK: till ReceiveMulti() is implemented, this
    //hack is required :)
    //here should be a call to ReceiveMulti(), remove when fixed
    //BEGIN HACK (still better than usleep())
    {
      fd_set fdset;
      struct timeval tv;
      int maxsock;

      FD_ZERO(&fdset);
      maxsock = 0;
      for (i = 0; i < BT_MAX_CLIENTS; i++) {
        if (!ttmpcon[i])
          continue;
        FD_SET(ttmpcon[i]->socket, &fdset);
        if (ttmpcon[i]->socket > maxsock)
          maxsock = ttmpcon[i]->socket;
      }
      tv.tv_sec = 0;
      tv.tv_usec = BTSR_RECEIVE_TIMEOUT;
      //sleep till something happens on a channel or
      //timeout expires
      select(maxsock + 1, &fdset, NULL, NULL, &tv);
    }
    for (i = 0; i < BT_MAX_CLIENTS; i++) {
      if (!ttmpcon[i])
        continue;
      //check for messages
      res = TTMPConnectionReceive(ttmpcon[i], &msg);
      if (res == TTMP_ERR_TIMEOUT)
        continue;
      if (res < TTMP_ERR_NONE) {
        DEBUG_LOG1("client has been removed with id %d", i);
        for (j = 0; j < BT_MAX_HANDLES; j++)
          if (refcon[j][i])
            refcon[j][i]--;
        TTMPConnectionDestroy(ttmpcon[i]);
        ttmpcon[i] = NULL;
        continue;
      }
      //handle the received message
      requesthandler(i, msg);
      TTMPMessageDestroy(msg);
    }
    //END HACK
    /* check if communication to clients was OK */
    for (i = 0; i < BT_MAX_CLIENTS; i++)
      if (ttmppend[i]) {
        DEBUG_LOG1("error happend during communication with a client id %d, removing the client", i);
        for (j = 0; j < BT_MAX_HANDLES; j++)
          if (refcon[j][i])
            refcon[j][i]--;
        TTMPConnectionDestroy(ttmpcon[i]);
        ttmpcon[i] = NULL;
        ttmppend[i] = 0;
      }
    /* cleanup unused handles */
    for (i = 0; i < BT_MAX_HANDLES; i++) {
      if (!btcon[i])
        continue;
      res = refhandle(i);
      if (!res) {
        res = BTHandlerGetState(btcon[i], &state);
        if (res != BT_ERR_NONE || state == BT_STATE_UNINITIALIZED) {
          DEBUG_LOG1("all clients left for the handle %d. Destroying...", i);
          BTHandlerDestroy(btcon[i]);
          btcon[i] = NULL;
        } else if (state != BT_STATE_CLOSING) {
          DEBUG_LOG1("all clients left for the handle %d. Closing...", i);
          BTHandlerClose(btcon[i]);
        }
      }
    }

#ifdef HANDLE_PAGESCAN
    /* try enable pagescan */
    /* PLEASECHECK: All pid stuff should be removed when pagescan enabling/disabling can be non-blocking */
    if (pagescandisablepid > 0)
      if (waitpid(pagescandisablepid, &res, WNOHANG))
        pagescandisablepid = -1;
    if (!pagescanenabled && pagescandisablepid < 0) {
      for (i = 0; i < BT_MAX_HANDLES; i++)
        if (btcon[i])
          if (BTHandlerGetState(btcon[i], &res) == BT_ERR_NONE)
            if (res == BT_STATE_START_LISTENING || res == BT_STATE_LISTENING)
              break;
      if (i != BT_MAX_HANDLES) {
        DEBUG_LOG("enabling pagescan.");
        pagescanenablepid = fork();
        if (pagescanenablepid == 0) {
          enablepagescan();
          exit(0);
        } else if (pagescanenablepid < 0) {
          DEBUG_LOG("enabling pagescan failed.");
        } else {
          pagescanenabled = 1;
        }
      }
    }
#endif /* HANDLE_PAGESCAN */

    /* try enable pairing */
    if (!pairingenabled) {
      for (i = 0; i < BT_MAX_HANDLES; i++)
        if (btcon[i])
          if (BTHandlerNeedPair(btcon[i], &res) == BT_ERR_NONE)
            if (res)
              break;
      if (i != BT_MAX_HANDLES) {
        DEBUG_LOG("enabling pairing.");
        BTSendSig(BT_HCID_NAME, BT_HCID_PAIR_ENABLE);
        pairingenabled = 1;
      }
    }
    /* run the active connections */
    for (i = 0; i < BT_MAX_HANDLES; i++) {
      if (btcon[i]) {
        /* walkaround, check if busy connecting, due to some bugs in phones and Bristol we need to serialize connection requests */
        disallow_connect = 0;
        for (j = 0; j < BT_MAX_HANDLES; j++) {
          if (btcon[j]) {
            BTHandlerGetState(btcon[j], &res);
            if ((res == BT_STATE_APP_START || res == BT_STATE_CONNECTING) && j != i) {
              disallow_connect = 1;
              break;
            }
          }
        }
        BTHandlerRun(btcon[i], disallow_connect);
      }
    }
    /* try disable pairing */
    if (pairingenabled) {
      for (i = 0; i < BT_MAX_HANDLES; i++)
        if (btcon[i])
          if (BTHandlerNeedPair(btcon[i], &res) == BT_ERR_NONE)
            if (res)
              break;
      if (i == BT_MAX_HANDLES) {
#ifndef BT_HID_REMOTE
        DEBUG_LOG("disabling pairing.");
        BTSendSig(BT_HCID_NAME, BT_HCID_PAIR_DISABLE);
        pairingenabled = 0;
#endif
      }
    }

#ifdef HANDLE_PAGESCAN
    /* try disable page scan */
    /* PLEASECHECK: All pid stuff should be removed when pagescan enabling/disabling can be non-blocking */
    if (pagescanenablepid > 0)
      if (waitpid(pagescanenablepid, &res, WNOHANG))
        pagescanenablepid = -1;
    if (pagescanenabled && pagescanenablepid < 0) {
      for (i = 0; i < BT_MAX_HANDLES; i++)
        if (btcon[i])
          if (BTHandlerGetState(btcon[i], &res) == BT_ERR_NONE)
            if (res == BT_STATE_START_LISTENING || res == BT_STATE_LISTENING)
              break;
      if (i == BT_MAX_HANDLES) {
        DEBUG_LOG("disabling pagescan.");
        pagescandisablepid = fork();
        if (pagescandisablepid == 0) {
          disablepagescan();
          exit(0);
        } else if (pagescandisablepid < 0) {
          DEBUG_LOG("disabling pagescan failed.");
        } else {
          pagescanenabled = 0;
        }
      }
    }
#endif /* HANDLE_PAGESCAN */
  }
  /* Close all the bluetooth connections */
  for (i = 0; i < BT_MAX_HANDLES; i++)
    if (btcon[i])
      BTHandlerClose(btcon[i]);
  /* run the active connections */
  for (i = 0; i < BT_MAX_HANDLES; i++)
    if (btcon[i]) {
      res = BTHandlerGetState(btcon[i], &state);
      while (state != BT_STATE_UNINITIALIZED) {
        BTHandlerRun(btcon[i], 1);
        usleep(BTSR_RECEIVE_TIMEOUT);
      }
    }
  /* destroy all the bluetooth connections */
  for (i = 0; i < BT_MAX_HANDLES; i++)
    if (btcon[i])
      BTHandlerDestroy(btcon[i]);
  /* destroy all the client connections */
  for (i = 0; i < BT_MAX_CLIENTS; i++)
    if (ttmpcon[i])
      TTMPConnectionDestroy(ttmpcon[i]);
  TTMPConnectionDestroy(newttmpcon);
  TTMPServerDestroy(svr);

  return 0;
}
